@@ -87,7 +87,7 @@ static const struct option_usage usage[] = {
 	{"port",         0, OPT_DEC32,  {.unum = 2222},
 		"Set TCP listening port (only used when operating in forward mode)\n"
 	},
-	/** --remote-adr */
+	/** --remote-addr */
 	{"address",      1, OPT_STR,    {.str = "127.0.0.1"},
 		"Set remote proxy destination address if client\n"
 		"Restrict to only this destination address if server\n"
@@ -113,6 +113,13 @@ static const struct option_usage usage[] = {
 #endif
 		"Enable libpcap on the given device.\n"
 	},
+	/** --list-libpcap-devices */
+	{NULL,           0, OPT_BOOL,   {.num = 0},
+#ifndef HAVE_PCAP
+		"(Not available on this platform.)\n"
+#endif
+		"List all available pcap devices.\n"
+	},
 	/** --logfile */
 	{"file",         0, OPT_STR,    {.str = "/var/log/ptunnel.log"},
 		"Specify a file to log to, rather than printing to standard out.\n"
@@ -136,6 +143,32 @@ static const struct option_usage usage[] = {
 		"Unprivileged mode will only work on some systems, and is in general less reliable\n"
 		"than running in privileged mode.\n"
 	},
+	/** --window-size */
+	{"packets",      0, OPT_DEC32,  {.unum = 64},
+		"Tune the number of packets that can be in-flight at the same time.\n"
+		"Increasing the window size will improve the maximum potential bandwidth.\n"
+	},
+	/** --ack-interval */
+	{"milliseconds", 0, OPT_DEC32,  {.unum = 1000},
+		"Tune the explicit acknowledgement interval (in milliseconds)\n"
+		"Decreasing the acknowledgement interval can improve NAT stability.\n"
+	},
+	/** --resend-interval */
+	{"milliseconds", 0, OPT_DEC32,  {.unum = 1500},
+		"Tune the lost packet timeout (in milliseconds)\n"
+		"Decreasing the resend interval can compensate for frequent packet loss.\n"
+	},
+	/** --payload-size */
+	{"bytes",        0, OPT_DEC32,  {.unum = 1024},
+		"Tune the amount of data per packet (in bytes)\n"
+		"Decreasing the payload size can avoid corruption of large packets.\n"
+		"Increasing the payload size can compensate for out-of-order delivery.\n"
+	},
+	/** --empty-pings */
+	{"count",        0, OPT_DEC32,  {.unum = 0},
+		"Tune the number of empty pings to send with each explicit acknowledgement.\n"
+		"Empty pings can compensate for ICMP sequence number inspection.\n"
+	},
 	/** --daemon */
 	{"pidfile",      0, OPT_STR,    {.str = "/run/ptunnel.pid"},
 #ifdef WIN32
@@ -188,16 +221,22 @@ static struct option long_options[] = {
 	{"magic",       required_argument, 0, 'm'},
 	{"proxy",       required_argument, 0, 'p'},
 	{"listen",      required_argument, 0, 'l'},
-	{"remote-adr",  optional_argument, 0, 'r'},
+	{"remote-addr", optional_argument, 0, 'r'},
 	{"remote-port", optional_argument, 0, 'R'},
 	{"connections", required_argument, 0, 'c'},
 	{"verbosity",   required_argument, 0, 'v'},
 	{"libpcap",     required_argument, 0, 'L'},
+	{"list-libpcap-devices", no_argument, &opts.list_pcap_devices, 1},
 	{"logfile",     optional_argument, 0, 'o'},
 	{"statistics",        no_argument, 0, 's'},
 	{"passwd",      required_argument, 0, 'P'},
-	{"udp",               no_argument, &opts.udp, 1 },
-	{"unprivileged",      no_argument, &opts.unprivileged, 1 },
+	{"udp",               no_argument, &opts.udp, 1},
+	{"unprivileged",      no_argument, &opts.unprivileged, 1},
+	{"window-size", required_argument, 0, 'w'},
+	{"ack-interval", required_argument, 0, 'a'},
+	{"resend-interval", required_argument, 0, 't'},
+	{"payload-size", required_argument, 0, 'y'},
+	{"empty-pings", required_argument, 0, 'E'},
 	{"daemon",      optional_argument, 0, 'd'},
 	{"syslog",            no_argument, 0, 'S'},
 	{"user",        optional_argument, 0, 'u'},
@@ -211,7 +250,7 @@ static struct option long_options[] = {
 
 static const void *get_default_optval(enum option_type opttype, const char *optname) {
 	for (unsigned i = 0; i < ARRAY_SIZE(long_options); ++i) {
-		if (strncmp(long_options[i].name, optname, strlen(long_options[i].name)) == 0) {
+		if (strncmp(long_options[i].name, optname, BUFSIZ /* not optimal */) == 0) {
 			assert(usage[i].otype == opttype);
 			return &usage[i].str;
 		}
@@ -231,7 +270,7 @@ static void set_options_defaults(void) {
 	opts.magic           = *(uint32_t *)  get_default_optval(OPT_HEX32, "magic");
 	opts.mode            = kMode_proxy;
 	opts.tcp_listen_port = *(uint32_t *)  get_default_optval(OPT_DEC32, "listen");
-	opts.given_dst_hostname = strdup(*(char **) get_default_optval(OPT_STR, "remote-adr"));
+	opts.given_dst_hostname = strdup(*(char **) get_default_optval(OPT_STR, "remote-addr"));
 	opts.given_dst_port  = *(uint32_t *)  get_default_optval(OPT_DEC32, "remote-port");
 	opts.max_tunnels     = *(uint32_t *)  get_default_optval(OPT_DEC32, "connections");
 	opts.log_level       = *(int *)       get_default_optval(OPT_DEC32, "verbosity");
@@ -241,8 +280,6 @@ static void set_options_defaults(void) {
 	opts.log_path        = strdup(*(char **)get_default_optval(OPT_STR, "logfile"));
 	opts.log_file        = stdout;
 	opts.print_stats     = *(int *)       get_default_optval(OPT_BOOL,  "statistics");
-	opts.udp             = *(int *)       get_default_optval(OPT_BOOL,  "udp");
-	opts.unprivileged    = *(int *)       get_default_optval(OPT_BOOL,  "unprivileged");
 #ifndef WIN32
 	opts.pid_path        = strdup(*(char **)get_default_optval(OPT_STR, "daemon"));
 
@@ -276,9 +313,9 @@ static void print_multiline(const char *prefix, const char *multiline) {
 	do {
 		if (start) {
 			end = strstr(start, sep);
-			if (end) {
+			if (end && *end != '\0') {
 				printf("%s%.*s\n", prefix, (int)(end-start), start);
-				start = end + strlen(sep);
+				start = end + strnlen(sep, BUFSIZ /* not optimal */);
 			}
 		}
 	} while (start && end);
@@ -380,7 +417,11 @@ int parse_options(int argc, char **argv) {
 
 	/* parse command line arguments */
 	while (1) {
-		c = getopt_long(argc, argv, "m:p:l:r::R::c:v:L::o::sP:d::Su::g::C::e::h", &long_options[0], &oidx);
+        /* FIXME: We are using '::' (optional argument values). This is not optimal
+         *        since you have to pass long options as '--option=value'. Commonly used
+         *        '--option value' is *NOT* allowed for some libc implementations.
+         */
+		c = getopt_long(argc, argv, "m:p:l:r::R::c:v:L::o::sP:d::Su::g::C::e::w:a:t:y:E:h", &long_options[0], &oidx);
 		if (c == -1) break;
 
 		switch (c) {
@@ -460,10 +501,10 @@ int parse_options(int argc, char **argv) {
 				pt_log(kLog_debug, "Password set - unauthenicated connections will be refused.\n");
 				//  Compute the password digest
 				md5_init(&state);
-				md5_append(&state, (md5_byte_t*)optarg, strlen(opts.password));
+				md5_append(&state, (md5_byte_t*)optarg, strnlen(opts.password, BUFSIZ /* not optimal */));
 				md5_finish(&state, &opts.password_digest[0]);
 				//  Hide the password in process listing
-				memset(optarg, '*', strlen(optarg));
+				memset(optarg, '*', strnlen(optarg, BUFSIZ /* not optimal */));
 				break;
 #ifndef WIN32
 			case 'd':
@@ -529,6 +570,31 @@ int parse_options(int argc, char **argv) {
 				pt_log(kLog_error, "SeLinux: %s\n", "feature not supported");
 				exit(1);
 #endif
+			case 'w':
+				if (!optarg)
+					break;
+				opts.window_size = atoi(optarg);
+				break;
+			case 'a':
+				if (!optarg)
+					break;
+				opts.ack_interval = atoi(optarg);
+				break;
+			case 't':
+				if (!optarg)
+					break;
+				opts.resend_interval = atoi(optarg);
+				break;
+			case 'y':
+				if (!optarg)
+					break;
+				opts.payload_size = atoi(optarg);
+				break;
+			case 'E':
+				if (!optarg)
+					break;
+				opts.empty_pings = atoi(optarg);
+				break;
 			case 'h':
 				print_usage(argv[0]);
 				exit(EXIT_SUCCESS);
@@ -539,7 +605,7 @@ int parse_options(int argc, char **argv) {
 	}
 
 	if (optind != argc) {
-		pt_log(kLog_error, "Unknown argument: %s\n", argv[optind]);
+		pt_log(kLog_error, "Unknown argument: '%s'\n", argv[optind]);
 		exit(1);
 	}
 
