@@ -82,7 +82,7 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 		                     sizeof(icmp_echo_packet_t) +
 		                     sizeof(ping_tunnel_pkt_t), bytes);
 	else {
-		if (opts.udp) {
+		if (opts.udp || opts.unprivileged) {
 			ip_pkt      = 0;
 			pkt         = (icmp_echo_packet_t*)buf;
 			pt_pkt      = (ping_tunnel_pkt_t*)pkt->data;
@@ -96,6 +96,7 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 		if (ntohl(pt_pkt->magic) == opts.magic) {
 			pt_pkt->state       = ntohl(pt_pkt->state);
 			pkt->identifier     = ntohs(pkt->identifier);
+			pkt->seq            = ntohs(pkt->seq);
 			pt_pkt->id_no       = ntohs(pt_pkt->id_no);
 			pt_pkt->seq_no      = ntohs(pt_pkt->seq_no);
 			/* Find the relevant connection, if it exists */
@@ -114,8 +115,10 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 			 */
 			if (cur) {
 				type_flag           = cur->type_flag;
-				if (type_flag == (uint32_t)kProxy_flag)
+				if (type_flag == (uint32_t)kProxy_flag) {
 					cur->icmp_id    = pkt->identifier;
+					cur->ping_seq   = pkt->seq;
+				}
 				if (!is_pcap)
 					cur->xfer.icmp_in++;
 			}
@@ -128,11 +131,13 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 				pt_log(kLog_error, "Dropping packet with invalid state.\n");
 				return;
 			}
-			pt_log(kLog_sendrecv, "Recv: %d [%d] bytes "
-			                      "[seq = %d] [type = %s] "
+			pt_log(kLog_sendrecv, "Recv: %4d [%4d] bytes "
+			                      "[id = 0x%04X] [seq = %d] "
+			                      "[seq_no = %d] [type = %s] "
 			                      "[ack = %d] [icmp = %d] "
-                                  "[user = %s] [pcap = %d]\n",
-                                  bytes, ntohl(pt_pkt->data_len),
+			                      "[user = %s] [pcap = %d]\n",
+			                      bytes, ntohl(pt_pkt->data_len),
+			                      pkt->identifier, ntohs(pkt->seq),
 			                      pt_pkt->seq_no, state_name[pt_pkt->state & (~kFlag_mask)],
 			                      ntohl(pt_pkt->ack), pkt->type,
 			                      (pkt_flag == kUser_flag ? "yes" : "no"), is_pcap);
@@ -179,6 +184,12 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 							pt_log(kLog_error, "Failed to create proxy descriptor!\n");
 							return;
 						}
+						if (pt_pkt->data_len > 0) {
+							handle_data(pkt, bytes, 0, 0, 0, 0, cur, 0);
+							if (!opts.password) {
+								handle_extended_options(cur);
+							}
+						}
 						if (init_state == kProto_authenticate) {
 							pt_log(kLog_debug, "Sending authentication challenge..\n");
 							/* Send challenge */
@@ -190,7 +201,7 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 							             &cur->send_idx, &cur->send_wait_ack, 0, 0,
 							             kProto_authenticate | cur->type_flag,
 							             &cur->dest_addr, cur->next_remote_seq,
-							             &cur->send_first_ack, &cur->ping_seq);
+							             &cur->send_first_ack, &cur->ping_seq, cur->window_size);
 						}
 					}
 					else if (type_flag == kUser_flag) {
@@ -231,13 +242,13 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 						             &cur->my_seq, cur->send_ring, &cur->send_idx,
 						             &cur->send_wait_ack, 0, 0,
 						             kProto_authenticate | cur->type_flag, &cur->dest_addr,
-						             cur->next_remote_seq, &cur->send_first_ack, &cur->  ping_seq);
+						             cur->next_remote_seq, &cur->send_first_ack, &cur->  ping_seq, cur->window_size);
 						/* We have authenticated locally.
 						 * It's up to the proxy now if it accepts our   response or not..
 						 */
 						cur->authenticated  = 1;
 						handle_data(pkt, bytes, cur->recv_ring, &cur->recv_wait_send,
-						            &cur->recv_idx, &cur->next_remote_seq);
+						            &cur->recv_idx, &cur->next_remote_seq, 0, cur->window_size);
 						return;
 					}
 					/* If proxy: Handle client's response to challenge */
@@ -247,6 +258,7 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 						                       cur->authenticated)
 						{
 							pt_log(kLog_verbose, "Remote end authenticated successfully.\n");
+							handle_extended_options(cur);
 							/* Authentication has succeeded, so now we can proceed
 							 * to handle incoming   TCP data.
 							 */
@@ -256,7 +268,7 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 							 * confusing the reliab  ility mechanism.
 							 */
 							handle_data(pkt, bytes, cur->recv_ring, &cur->recv_wait_send,
-							            &cur->recv_idx, &cur->next_remote_seq);
+							            &cur->recv_idx, &cur->next_remote_seq, 0, cur->window_size);
 						}
 						else {
 							pt_log(kLog_info, "Remote end failed authentication.\n");
@@ -285,16 +297,23 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
 				}
 
 				if (cur && cur->sock) {
+					double now = time_as_double();
+					if (pt_pkt->state != kProto_ack) {
+						cur->last_data_activity = now;
+					}
 					if (pt_pkt->state == kProto_data || pt_pkt->state == kProxy_start ||
 					    pt_pkt->state   == kProto_ack)
 					{
+						if (pt_pkt->state == kProxy_start) {
+							pt_pkt->data_len = 0;
+						}
 						handle_data(pkt, bytes, cur->recv_ring, &cur->recv_wait_send,
-						            &cur->recv_idx,   &cur->next_remote_seq);
+						            &cur->recv_idx,   &cur->next_remote_seq, 0, cur->window_size);
 					}
 					handle_ack((uint16_t)pt_pkt->ack, cur->send_ring, &cur->send_wait_ack,
 					           0, cur->send_idx, &cur->send_first_ack, &cur->remote_ack_val,
-					           is_pcap);
-					cur->last_activity      = time_as_double();
+					           is_pcap, cur->window_size);
+					cur->last_activity = now;
 				}
 			}
 		}
@@ -308,7 +327,7 @@ void handle_packet(char *buf, unsigned bytes, int is_pcap, struct sockaddr_in *a
  * onto the passed-in receive ring.
  */
 void handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[],
-                 int *await_send, int *insert_idx,  uint16_t *next_expected_seq)
+                 int *await_send, int *insert_idx,  uint16_t *next_expected_seq, void *vcur, uint16_t window_size)
 {
 	ping_tunnel_pkt_t *pt_pkt      = (ping_tunnel_pkt_t*)pkt->data;
 	int               expected_len = sizeof(ip_packet_t) + sizeof(icmp_echo_packet_t) +
@@ -326,7 +345,7 @@ void handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[],
 	 */
 	expected_len     += pt_pkt->data_len;
 	expected_len     += expected_len % 2;
-	if (opts.udp)
+	if (opts.udp || opts.unprivileged)
 		expected_len -= sizeof(ip_packet_t);
 	if (total_len < expected_len) {
 		pt_log(kLog_error, "Packet not completely received: %d Should be: %d. "
@@ -337,6 +356,25 @@ void handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[],
 		 */
 		exit(0);
 	}
+	if (vcur) {
+		proxy_desc_t *cur = (proxy_desc_t *)vcur;
+		uint16_t *extended_options = (uint16_t *)pt_pkt->data;
+		if (pt_pkt->data_len >= 2) {
+			cur->extended_options[0] = ntohs(extended_options[0]);
+		}
+		if (pt_pkt->data_len >= 4) {
+			cur->extended_options[1] = ntohs(extended_options[1]);
+		}
+		if (pt_pkt->data_len >= 6) {
+			cur->extended_options[2] = ntohs(extended_options[2]);
+		}
+		if (pt_pkt->data_len >= 8) {
+			cur->extended_options[3] = ntohs(extended_options[3]);
+		}
+		return;
+	}
+	if (!next_expected_seq)
+		return;
 	if (pt_pkt->seq_no == *next_expected_seq) {
 		/* hmm, what happens if this test is true? */
 		if (!ring[*insert_idx]) { /* && pt_pkt->state == kProto_data */
@@ -349,14 +387,14 @@ void handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[],
 			pt_log(kLog_debug, "Dup packet?\n");
 
 		(*next_expected_seq)++;
-		if (*insert_idx >= kPing_window_size)
+		if (*insert_idx >= window_size)
 			*insert_idx	= 0;
 		/* Check if we have already received some of the next packets */
 		while (ring[*insert_idx]) {
 			if (ring[*insert_idx]->seq_no == *next_expected_seq) {
 				(*next_expected_seq)++;
 				(*insert_idx)++;
-				if (*insert_idx >= kPing_window_size)
+				if (*insert_idx >= window_size)
 					*insert_idx	= 0;
 			}
 			else
@@ -371,13 +409,13 @@ void handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[],
 		d   = s - r;
 		if (d < 0) { /* This packet _may_ be old, or seq_no may have wrapped around */
 			d = (s+0xFFFF) - r;
-			if (d < kPing_window_size) {
+			if (window_size && d < window_size) {
 				/* Counter has wrapped, so we should add this packet to the recv ring */
-				pos	= ((*insert_idx)+d) % kPing_window_size;
+				pos	= ((*insert_idx)+d) % window_size;
 			}
 		}
-		else if (d < kPing_window_size)
-			pos	= ((*insert_idx)+d) % kPing_window_size;
+		else if (window_size && d < window_size)
+			pos	= ((*insert_idx)+d) % window_size;
 
 		if (pos != -1) {
 			if (!ring[pos]) {
@@ -394,16 +432,44 @@ void handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[],
 	}
 }
 
+void handle_extended_options(void *vcur)
+{
+	proxy_desc_t *cur = (proxy_desc_t *)vcur;
+	if (cur->extended_options[0] > 0) {
+		if (cur->extended_options[0] > cur->window_size) {
+			size_t extend = cur->extended_options[0] - cur->window_size;
+			cur->send_ring = realloc(cur->send_ring, cur->extended_options[0] * sizeof(icmp_desc_t));
+			cur->recv_ring = realloc(cur->recv_ring, cur->extended_options[0] * sizeof(forward_desc_t *));
+			memset(cur->send_ring + cur->window_size, 0, extend * sizeof(icmp_desc_t));
+			memset(cur->recv_ring + cur->window_size, 0, extend * sizeof(forward_desc_t *));
+		}
+		cur->window_size = cur->extended_options[0];
+		pt_log(kLog_verbose, "Received extended option for window size %d \n", cur->window_size);
+	}
+	if (cur->extended_options[1] > 0) {
+		cur->ack_interval = cur->extended_options[1] / 1000.0;
+		pt_log(kLog_verbose, "Received extended option for ack interval %f \n", cur->ack_interval);
+	}
+	if (cur->extended_options[2] > 0) {
+		cur->resend_interval = cur->extended_options[2] / 1000.0;
+		pt_log(kLog_verbose, "Received extended option for resend interval %f \n", cur->resend_interval);
+	}
+	if (cur->extended_options[3] > 0) {
+		cur->payload_size = cur->extended_options[3];
+		pt_log(kLog_verbose, "Received extended option for payload size %d \n", cur->payload_size);
+	}
+}
+
 void handle_ack(uint16_t seq_no, icmp_desc_t ring[], int *packets_awaiting_ack,
                 int one_ack_only, int insert_idx, int *first_ack,
-                uint16_t *remote_ack, int is_pcap)
+                uint16_t *remote_ack, int is_pcap, uint16_t window_size)
 {
 	int i, j, k;
 	ping_tunnel_pkt_t *pt_pkt;
 
 	if (*packets_awaiting_ack > 0) {
 		if (one_ack_only) {
-			for (i = 0; i < kPing_window_size; i++) {
+			for (i = 0; i < window_size; i++) {
 				if (ring[i].pkt && ring[i].seq_no == seq_no && !is_pcap) {
 					pt_log(kLog_debug, "Received ack for only seq %d\n", seq_no);
 					pt_pkt	= (ping_tunnel_pkt_t*)ring[i].pkt->data;
@@ -411,10 +477,11 @@ void handle_ack(uint16_t seq_no, icmp_desc_t ring[], int *packets_awaiting_ack,
 					*remote_ack	= (uint16_t)ntohl(pt_pkt->ack);
 					free(ring[i].pkt);
 					ring[i].pkt	= 0;
+					ring[i].pkt_len	= 0;
 					(*packets_awaiting_ack)--;
 					if (i == *first_ack) {
-						for (j=1;j<kPing_window_size;j++) {
-							k	= (i+j)%kPing_window_size;
+						for (j=1;j<window_size;j++) {
+							k	= (i+j)%window_size;
 							if (ring[k].pkt) {
 								*first_ack = k;
 								break;
@@ -433,10 +500,10 @@ void handle_ack(uint16_t seq_no, icmp_desc_t ring[], int *packets_awaiting_ack,
 			int	i, can_ack = 0, count = 0;
 			i	= insert_idx-1;
 			if (i < 0)
-				i	= kPing_window_size - 1;
+				i	= window_size - 1;
 
 			pt_log(kLog_debug, "Received ack-series starting at seq %d\n", seq_no);
-			while (count < kPing_window_size) {
+			while (count < window_size) {
 				if (!ring[i].pkt)
 					break;
 
@@ -448,11 +515,12 @@ void handle_ack(uint16_t seq_no, icmp_desc_t ring[], int *packets_awaiting_ack,
 				if (can_ack) {
 					free(ring[i].pkt);
 					ring[i].pkt	= 0;
+					ring[i].pkt_len	= 0;
 					(*packets_awaiting_ack)--;
 				}
 				i--;
 				if (i < 0)
-					i	= kPing_window_size - 1;
+					i	= window_size - 1;
 				count++;
 			}
 		}
