@@ -34,7 +34,7 @@
  * POSSIBILITY OF SUCH DAMAGE.
  *
  * Contacting the author:
- * You can get in touch with me, Daniel Stdle (that's the Norwegian letter oe,
+ * You can get in touch with me, Daniel Stoedle (that's the Norwegian letter oe,
  * in case your text editor didn't realize), here: <daniels@cs.uit.no>
  *
  * The official ptunnel website is here:
@@ -88,7 +88,6 @@ uint32_t num_tunnels = 0;
 uint32_t *seq_expiry_tbl = NULL;
 
 /* Some buffer constants */
-const int tcp_receive_buf_len  = kDefault_buf_size;
 const int icmp_receive_buf_len = kDefault_buf_size + kIP_header_size +
                                  kICMP_header_size + sizeof(ping_tunnel_pkt_t);
 const int pcap_buf_size        = (kDefault_buf_size + kIP_header_size +
@@ -98,9 +97,34 @@ char pcap_filter_program[]     = "icmp";
 
 /** The chain of client/proxy connections */
 proxy_desc_t *chain = 0;
-const char *state_name[kNum_proto_types] = { "start", "ack", "data",
+const char *state_name[kNum_proto_types] = { "start", "ack  ", "data ",
                                              "close", "authenticate" };
 
+#ifdef HAVE_PCAP
+static void print_pcap_devices(void) {
+	pcap_if_t *devs, *cur_dev;
+	pcap_addr_t *cur_addr;
+	char errbuf[PCAP_ERRBUF_SIZE+1];
+
+	if (pcap_findalldevs(&devs, errbuf)) {
+		pt_log(kLog_error, "List all available pcap devices failed: %s.\n", errbuf);
+	}
+	printf("Available pcap devices:\n");
+	for (cur_dev = devs; cur_dev; cur_dev = cur_dev->next) {
+		if (cur_dev->description)
+			printf("\n\t%s%c '%s'\n", cur_dev->name, (cur_dev->addresses ? ':' : ' '),
+				cur_dev->description);
+		else
+			printf("\n\t%s%c\n", cur_dev->name, (cur_dev->addresses ? ':' : ' '));
+		for (cur_addr = cur_dev->addresses; cur_addr; cur_addr = cur_addr->next) {
+			if (cur_addr->addr->sa_family == AF_INET)
+				printf("\t\t%s\n", inet_ntoa(((struct sockaddr_in*)cur_addr->addr)->sin_addr));
+		}
+	}
+	pcap_freealldevs(devs);
+}
+#endif
+
 /* Let the fun begin! */
 int main(int argc, char *argv[]) {
 #ifndef WIN32
@@ -137,6 +161,19 @@ int main(int argc, char *argv[]) {
 	if (parse_options(argc, argv))
 		return -1;
 
+	/* Init ptunnel RNG */
+	pt_random();
+
+	if (opts.list_pcap_devices) {
+#ifdef HAVE_PCAP
+		print_pcap_devices();
+		return 0;
+#else
+		pt_log(kLog_error, "Pcap not available!\n");
+		return 1;
+#endif
+	}
+
 #ifdef HAVE_PCAP
 	if (opts.pcap && opts.udp) {
 		pt_log(kLog_error, "Packet capture is not supported (or needed) when using UDP for transport.\n");
@@ -144,7 +181,10 @@ int main(int argc, char *argv[]) {
 	}
 #ifdef WIN32
 	if (!opts.pcap && !opts.udp) {
-		pt_log(kLog_info, "WARNING: Running ptunnel-ng on Windows in ICMP mode without WinPcap enabled is not supported and may not work!\n");
+		pt_log(kLog_info, "Running ptunnel-ng on Windows in ICMP mode without WinPcap enabled is not supported and may not work!\n");
+        pt_log(kLog_info, "If you encounter problems, install WinPCAP from:\n");
+		pt_log(kLog_info, "https://www.winpcap.org/install/default.htm or for WIN10: https://nmap.org/npcap/windows-10.html\n");
+		pt_log(kLog_info, "After WinPCAP is installed, you can list pcap devices with: --list-pcap-devices\n");
 	}
 #endif
 #endif
@@ -179,7 +219,7 @@ int main(int argc, char *argv[]) {
 	}
 	if (opts.chroot) {
 		pt_log(kLog_info, "Restricting file access to %s\n", opts.root_dir);
-		if (-1 == chdir(opts.root_dir) || -1 == chroot(opts.root_dir)) {
+		if (-1 == chdir(opts.root_dir) || -1 == chroot(".") || -1 == chdir("/")) {
 			pt_log(kLog_error, "chdir/chroot `%s': %s\n", opts.root_dir, strerror(errno));
 			exit(1);
 		}
@@ -532,7 +572,13 @@ void* pt_proxy(void *args) {
 		max_sock = fwd_sock+1;
 		pthread_mutex_lock(&chain_lock);
 		for (cur = chain; cur; cur = cur->next) {
-			if (cur->sock) {
+			/* Only handle traffic if there is traffic on the socket, we have
+			 * room in our send window AND we either don't use a password, or
+			 * have been authenticated.
+			 */
+			if (cur->sock && cur->send_wait_ack < cur->window_size &&
+			    (!opts.password || cur->authenticated))
+			{
 				FD_SET(cur->sock, &set);
 				if (cur->sock >= max_sock)
 					max_sock = cur->sock+1;
@@ -552,10 +598,29 @@ void* pt_proxy(void *args) {
 			if (cur->state == kProxy_start) {
 				pt_log(kLog_verbose, "Sending proxy request.\n");
 				cur->last_ack = time_as_double();
-				queue_packet(fwd_sock, cur->pkt_type, 0, 0, cur->id_no, cur->id_no,
+				uint16_t extended_options[4];
+				size_t extended_options_size = 0;
+				memset(extended_options, 0, sizeof(extended_options));
+				if (opts.window_size) {
+					extended_options_size = sizeof(uint16_t);
+					extended_options[0] = htons(opts.window_size);
+				}
+				if (opts.ack_interval) {
+					extended_options_size = 2*sizeof(uint16_t);
+					extended_options[1] = htons(opts.ack_interval);
+				}
+				if (opts.resend_interval) {
+					extended_options_size = 3*sizeof(uint16_t);
+					extended_options[2] = htons(opts.resend_interval);
+				}
+				if (opts.payload_size) {
+					extended_options_size = 4*sizeof(uint16_t);
+					extended_options[3] = htons(opts.payload_size);
+				}
+				queue_packet(fwd_sock, cur->pkt_type, (char *)extended_options, extended_options_size, cur->id_no, cur->id_no,
 				             &cur->my_seq, cur->send_ring, &cur->send_idx, &cur->send_wait_ack,
 				             cur->dst_ip, cur->dst_port, cur->state | cur->type_flag,
-				             &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
+				             &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq, cur->window_size);
 				cur->xfer.icmp_out++;
 				cur->state = kProto_data;
 			}
@@ -567,14 +632,9 @@ void* pt_proxy(void *args) {
 				remove_proxy_desc(cur, prev);
 				continue;
 			}
-			/* Only handle traffic if there is traffic on the socket, we have
-			 * room in our send window AND we either don't use a password, or
-			 * have been authenticated.
-			 */
-			if (FD_ISSET(cur->sock, &set) && cur->send_wait_ack < kPing_window_size &&
-			    (!opts.password || cur->authenticated))
-			{
-				bytes = recv(cur->sock, cur->buf, tcp_receive_buf_len, 0);
+			/* Handle TCP traffic */
+			if (FD_ISSET(cur->sock, &set)) {
+				bytes = recv(cur->sock, cur->buf, cur->payload_size, 0);
 				if (bytes <= 0) {
 					pt_log(kLog_info, "Connection closed or lost.\n");
 					tmp	= cur->next;
@@ -590,7 +650,7 @@ void* pt_proxy(void *args) {
 				queue_packet(fwd_sock, cur->pkt_type, cur->buf, bytes, cur->id_no,
 				             cur->icmp_id, &cur->my_seq, cur->send_ring, &cur->send_idx,
 				             &cur->send_wait_ack, 0, 0, cur->state | cur->type_flag,
-				             &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
+				             &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq, cur->window_size);
 			}
 			prev = cur;
 			tmp  = cur->next;
@@ -621,13 +681,14 @@ void* pt_proxy(void *args) {
 				continue;
 			}
 			if (cur->recv_wait_send && cur->sock)
-				cur->xfer.bytes_in += send_packets(cur->recv_ring, &cur->recv_xfer_idx, &cur->recv_wait_send, &cur->sock);
+				cur->xfer.bytes_in += send_packets(cur->recv_ring, &cur->recv_xfer_idx, &cur->recv_wait_send, &cur->sock, cur->window_size);
 
 			/* Check for any icmp packets requiring resend, and resend _only_ the first packet. */
 			idx	= cur->send_first_ack;
-			if (cur->send_ring[idx].pkt && cur->send_ring[idx].last_resend+kResend_interval < now) {
+			if (cur->send_ring[idx].pkt && cur->send_ring[idx].last_resend+cur->resend_interval < now) {
 				pt_log(kLog_debug, "Resending packet with seq-no %d.\n", cur->send_ring[idx].seq_no);
 				cur->send_ring[idx].last_resend   = now;
+				cur->send_ring[idx].pkt->identifier = htons(cur->icmp_id);
 				cur->send_ring[idx].pkt->seq      = htons(cur->ping_seq);
 				cur->ping_seq++;
 				cur->send_ring[idx].pkt->checksum = 0;
@@ -638,15 +699,29 @@ void* pt_proxy(void *args) {
 				cur->xfer.icmp_resent++;
 			}
 			/* Figure out if it's time to send an explicit acknowledgement */
-			if (cur->last_ack+1.0 < now && cur->send_wait_ack < kPing_window_size &&
+			if (cur->last_ack+cur->ack_interval < now && cur->send_wait_ack < cur->window_size &&
 			    cur->remote_ack_val+1 != cur->next_remote_seq)
 			{
-				cur->last_ack = now;
+				idx = cur->send_idx;
 				queue_packet(fwd_sock, cur->pkt_type, 0, 0, cur->id_no, cur->icmp_id,
 				             &cur->my_seq, cur->send_ring, &cur->send_idx, &cur->send_wait_ack,
 				             cur->dst_ip, cur->dst_port, kProto_ack | cur->type_flag,
-				             &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq);
+				             &cur->dest_addr, cur->next_remote_seq, &cur->send_first_ack, &cur->ping_seq, cur->window_size);
 				cur->xfer.icmp_ack_out++;
+				if (opts.empty_pings &&
+					cur->last_data_activity > cur->last_ack &&
+					cur->send_ring[idx].pkt_len > sizeof(icmp_echo_packet_t) &&
+					cur->send_ring[idx].pkt->type == kICMP_echo_request) {
+					for (uint16_t e = 0; e < opts.empty_pings; e++) {
+						cur->send_ring[idx].pkt->seq      = htons(cur->ping_seq);
+						cur->ping_seq++;
+						cur->send_ring[idx].pkt->checksum = 0;
+						cur->send_ring[idx].pkt->checksum = htons(calc_icmp_checksum((uint16_t*)cur->send_ring[idx].pkt, sizeof(icmp_echo_packet_t)));
+						sendto(fwd_sock, (const void*)cur->send_ring[idx].pkt, sizeof(icmp_echo_packet_t),
+						       0, (struct sockaddr*)&cur->dest_addr, sizeof(struct sockaddr));
+					}
+				}
+				cur->last_ack = now;
 			}
 		}
 		pthread_mutex_unlock(&chain_lock);
@@ -793,10 +868,10 @@ void send_termination_msg(proxy_desc_t *cur, int icmp_sock) {
 	queue_packet(icmp_sock, cur->pkt_type, 0, 0, cur->id_no, cur->icmp_id, &cur->my_seq,
 	             cur->send_ring, &cur->send_idx, &cur->send_wait_ack, 0, 0,
 	             kProto_close | cur->type_flag, &cur->dest_addr, cur->next_remote_seq,
-	             &cur->send_first_ack, &cur->ping_seq);
+	             &cur->send_first_ack, &cur->ping_seq, cur->window_size);
 	queue_packet(icmp_sock, cur->pkt_type, 0, 0, cur->id_no, cur->icmp_id, &cur->my_seq,
 	             cur->send_ring, &cur->send_idx, &cur->send_wait_ack, 0, 0,
 	             kProto_close | cur->type_flag, &cur->dest_addr, cur->next_remote_seq,
-	             &cur->send_first_ack, &cur->ping_seq);
+	             &cur->send_first_ack, &cur->ping_seq, cur->window_size);
 	cur->xfer.icmp_out += 2;
 }
